from .utils import DefaultLambda, DictUpdater
from .exceptions import LinkerExists, LinkerNotFound, LinkerNotReady

from typing import Any, Dict

class Linker:
    '''Linker object to link between attributes of a class' instance using property'''
    def __init__(self, sourceVar: str, getterConverter: callable = DefaultLambda, setterConverter: callable = DefaultLambda, setterOverrider: callable = None, doc: str = None):
        '''
        Params
        ------
        sourceVar:str is the source instance variable name
        getterConverter:Callable is the function that gets called to process the sourceVar's value. Called with a single argument, which is the instance self.
        setterConverter:Callable is the function that gets called to process a replacement before being set to the sourceVar. Called with 2 args=(instance_self, replacement)
        setterOverrider:Callable is the function that gets called to do the setter process. Called with 3 args=(instance_self, Linker.sourceVar, replacement)
        doc:str is the documentation string (docstring) for the property
        '''
        self.sourceVar = sourceVar
        self.getterConverter = getterConverter
        self.setterOverrider = setterOverrider
        self.setterConverter = setterConverter
        self.doc = doc or "Linker to instance variable: {}".format(self.sourceVar)
        self.property = None
        self.links = {} # dictionary of targetClass and targetVar, Ex: {Class1: [Var1,Var2,Var3,...],...}

    def __repr__(self):
        return "<{} SourceVar={} Ready={}>".format(self.__class__.__name__, self.sourceVar, self.ready)

    def __str__(self):
        return "{} SourceVariable={} ReadyToBeApplied={}".format(self.__class__.__name__, self.sourceVar, self.ready)

    @property
    def ready(self):
        '''Whether the linker has been set up or not.'''
        return isinstance(self.property, property)

    def setup(self, enableSetter: bool = True):
        '''
        Sets up the linker's property. This method needs to be called before applying.
        
        Params
        ------
        enableSetter:bool whether to enable setter for the property or not. Basically a full-access-link(True) or a read-only-link(False). Defaults to True(full access link)
        '''
        # linked self here is an instance of an object, no longer a class
        def _linkGetter(linkedSelf):
            return self.getterConverter(linkedSelf.__getattribute__(self.sourceVar))
        _linkGetter.linker = self

        if self.setterOverrider is None:
            def _linkSetter(linkedSelf,replacement):
                setattr(linkedSelf, self.sourceVar, self.setterConverter(linkedSelf, replacement))
        else:
            def _linkSetter(linkedSelf,replacement):
                self.setterOverrider(linkedSelf, self.sourceVar, replacement) # args=(targetInstanceSelf, linkedVariable, replacement)
        
        _linkSetter.linker = self
        # print(f"{enableSetter=}") # DEBUG
        self.property = property(fget=_linkGetter, fset=_linkSetter if enableSetter else None, doc=self.doc)
        return self # Support calling chain. Ex: Linker(...).setup().apply(...)

    def apply(self, targetClass: type, targetVar: str):
        '''
        Applies the linker to given class at the target variable name.

        Params
        ------
        targetClass:type is the class targeted to be linked with
        targetVar:str is the class's instance attribute name to be linked at
        '''
        if not self.ready:
            raise LinkerNotReady("Linker is not ready yet! Please call setup first to set it up.")
        setattr(targetClass, targetVar, self.property)
        self.links[targetClass] = self.links.get(targetClass, []) + [targetVar] #Ex: {Class1: [Var1,Var2,...]}


class LinkManager:
    _DEFAULT_MANAGER = None # Manager for preset linkers/bindings
    _MANAGERS = [] # Keep track of created managers
    _LINKER_CLASS = Linker # Linker class to create links

    def __new__(cls):
        # Keep track of managers
        instance = super().__new__(cls)
        cls._MANAGERS.append(instance)
        return instance

    @classmethod
    def _getDefault(cls):
        '''Gets the default link manager, if not available, create one. Used by presets.'''
        if not isinstance(cls._DEFAULT_MANAGER, cls):
            cls._DEFAULT_MANAGER = cls()
        return cls._DEFAULT_MANAGER

    @classmethod
    def changeLinkerClass(cls, linkerClass: type):
        '''
        Change Linker Class if needed, or for experimental purposes. Given Linker Class replacement MUST inherit from the original Linker Class.

        Params
        ------
        linkerClass:type is the replacement for LinkManager's linker class.
        '''
        if not issubclass(linkerClass, Linker):
            raise RuntimeError('Given Linker Class Replacement Must Inherit From <Linker>.')
        cls._LINKER_CLASS = linkerClass

    def __init__(self, autoLinkWithManager: bool = True, linkerSetupOptions: Dict[str, Any] = {}):
        '''
        Params
        ------
        autoLinkWithManager:bool is whether to record the link pairs in the LinkManager's links attribute
        linkerSetupOpntions:dict is to update the default setup options of the LinkManager'''
        self.autoLinkWithManager = autoLinkWithManager
        self.linkerSetupOptions = DictUpdater({'enableSetter':True}, linkerSetupOptions)
        self.linkers = {}
        self.links = {}

    def __repr__(self):
        return "<{} LinkerClass={} LinkersCount={}>".format(self.__class__.__name__, self.linkerClass, len(self.linkers))

    @property
    def linkerClass(self):
        return self.__class__._LINKER_CLASS

    def createLinker(self, name: str, sourceVar: str, getterConverter: callable = DefaultLambda, setterConverter: callable = DefaultLambda, setterOverrider: callable = None, doc: str = '', overwrite: bool = False, doSetup: bool = True, setupOptions: Dict[str, Any] = {}):
        '''
        Creates a linker object with given arguments, and put it into a hashmap on the LinkManager linkers attribute, then call its setup method if doSetup is True.

        Params
        ------
        name:str the name to be paired with the linker in the Manager's hashmap
        sourceVar:str the source attribute name of an instance for the linker to later be applied to
        getterConverter:Callable is the function that gets called to process the sourceVar's value
        setterConverter:Callable is the function that gets called to process a replacement before being set to the sourceVar
        setterOverrider:Callable is the function that gets called to do the setter process
        doc:str is the documentation string (docstring) for the property
        overwrite:bool whether to overwrite existing entries if found with the same name on the hashmap
        doSetup:bool whether to call the setup method of the linker after its creation
        setupOptions:dict extra setup options for the linker.setup method in the form of a dictionary
        '''
        if not overwrite and self.linkers.get(name) is not None:
            raise LinkerExists("A Linker with name '{}' already exists in this linker manager. To overwrite it, please set overwrite=True.".format(name))
        self.linkers[name] = self.linkerClass(sourceVar, getterConverter=getterConverter, setterConverter=setterConverter, setterOverrider=setterOverrider, doc=doc)
        if doSetup:
            setupOptions = DictUpdater(self.linkerSetupOptions, setupOptions)
            self.linkers[name].setup(**setupOptions)
        return self.linkers[name] # if they wanted to use it through call chaining, they can. Ex: AttrLinkManager().createLinker(...).apply(...)

    def prepareLinkers(self, setupOptions: Dict[str, Any] = {}):
        '''
        Calls the setup method for every linker in the manager's hashmap.
        
        Params
        ------
        setupOptions:dict extra setup options for the linker.setup method in the form of a dictionary
        '''
        # Setup linkers
        setupOptions = DictUpdater(self.linkerSetupOptions, setupOptions)
        [linker.setup(**setupOptions) for linker in self.linkers]

    def applyLinker(self, linkerName: str, targetClass: type, targetVar: str):
        '''
        Applies the linker with given linkerName to given class at the target variable name.

        Params
        ------
        linkerName:str is the name of the linker in the manager's hashmap
        targetClass:type is the class targeted to be linked with
        targetVar:str is the class's instance attribute name to be linked at
        '''
        try:
            self.linkers[linkerName].apply(targetClass, targetVar)
            if self.autoLinkWithManager:
                self.links[targetClass] = (targetVar, self.linkers[linkerName])
        except KeyError as exc:
            raise LinkerNotFound('Linker {} is not found in the manager. Make sure you enter the correct name, or create one if it does not exists.'.format(linkerName)) from exc

    def bind(self, targetClass: type, sourceVar: str, targetVar: str, getterConverter: callable = DefaultLambda, setterConverter: callable = DefaultLambda, setterOverrider: callable = None, doc: str = '', setupOptions: Dict[str, Any] = {}, name: str = None, orphan: bool = False, **kw):
        '''
        Creates linker and apply it to the targetClass, then store it with generated name or given name, or orphan it.

        Params
        ------
        targetClass:type is the class targeted to be linked with
        sourceVar:str the source attribute name of an instance for the linker to later be applied to
        targetVar:str is the class's instance attribute name to be linked at
        getterConverter:Callable is the function that gets called to process the sourceVar's value
        setterConverter:Callable is the function that gets called to process a replacement before being set to the sourceVar
        setterOverrider:Callable is the function that gets called to do the setter process
        doc:str is the documentation string (docstring) for the property
        setupOptions:dict extra setup options for the linker.setup method in the form of a dictionary
        name:str the name to be paired with the linker in the Manager's hashmap
        orphan:bool whether to orphan the linker after linking. Meaning leaving no bindings between the linker and manager. Defaults to False.

        Extra keyword argument passed, would be passed directly to manager.createLinker (Only if orphan is disabled)
        '''
        # generates name for the linker, or just make one and orphan/delete it after binding.
        setupOptions = DictUpdater(self.linkerSetupOptions, setupOptions)
        # print(setupOptions) # DEBUG
        if orphan:
            linker = self.linkerClass(sourceVar, getterConverter=getterConverter, setterConverter=setterConverter, setterOverrider=setterOverrider, doc=doc)
            linker.setup(**setupOptions)
            linker.apply(targetClass, targetVar)
            # del linker
            return linker
        
        # Generate a name if not given, though not nice looking, it should not be possible for any duplicates under normal usage, 
        # where each targetVar for each class is used once only.
        name = str(name or '{}-class:{};source:{};target:{}'.format('%s(%s)' % (self.linkerClass.__name__, id(self.linkerClass)), 
                                                                    '%s(%s)' % (targetClass.__name__, id(targetClass)), 
                                                                    sourceVar, targetVar))
        self.createLinker(name, sourceVar, getterConverter=getterConverter, setterConverter=setterConverter, setterOverrider=setterOverrider, doc=doc, setupOptions=setupOptions, **kw)
        self.applyLinker(name, targetClass, targetVar)




class PropBinder:
    '''
    PropBinder behaves similiarly to the whole attrLinker, but, PropBinder makes all instance of the applied/binded class to a specific instance's value.
    
    PropBinder has the advantage of being bindable to other object, without connection to the object going to be binded. 
    Unlike Linker, which can only bind to object that is its attribute.

    PropBinder might be of use to apps which only allow up to 1 instance of a class.
    '''
    def __init__(self, source_obj):
        self.source_obj = source_obj
        self.propList = {}
    
    def __repr__(self):
        return "<{} Source={}>".format(self.__class__.__name__, self.source_obj)
    
    def createProp(self, converter=lambda x:x):
        def _func(_self):
            return converter(self.source_obj)
        return property(_func)
    
    def applyProp(self, prop, target_class, target_name):
        setattr(target_class, target_name, prop)
        self.propList["{}.{}".format(target_class, target_name)] = prop
    
    def bind(self, target_class, target_name, converter=lambda x:x):
        prop = self.createProp(converter)
        return self.applyProp(prop, target_class, target_name)
        




from .attrLinker import LinkManager
from .linkMethod import LinkMethod


class PreparedLink:
    def __init__(self, linkMethod: LinkMethod, sourceVar, *args, **kwargs):
        self.linkMethod = linkMethod
        self.sourceVar = sourceVar
        self.executionArgs = args
        self.executionKwargs = kwargs
        
        self._appliedClasses = []
    
    def __repr__(self):
        return "<{} Method={} AppliedCount={}>".format(self.__class__.__name__, LinkMethod(self.linkMethod).name, len(self._appliedClasses))

    def apply(self, targetClass: type):
        if self.linkMethod is None:
            manager = LinkManager._getDefault()
            manager.bind(targetClass, self.sourceVar, *self.executionArgs, **self.executionKwargs)
        else:
            self.linkMethod(targetClass, self.sourceVar, *self.executionArgs, **self.executionKwargs)
        self._appliedClasses.append(targetClass)

    @classmethod
    def applyLinks(targetClass, links:list):
        [link.apply(targetClass) for link in links]



from enum import Enum

from .presets import linkDictionary, multiLinkDictionary, linkList, multiLinkList, linkObject, multiLinkObject, formattedTextFromDict


class LinkMethod(Enum):
    DirectLink = None # A bypass to manager.bind
    Dictionary = linkDictionary
    MultiDictionary = multiLinkDictionary
    List = linkList
    MultiList = multiLinkList
    Object = linkObject
    MultiObject = multiLinkObject
    FormattedText = formattedTextFromDict

DirectLink = LinkMethod.DirectLink
Dictionary = LinkMethod.Dictionary
MultiDictionary = LinkMethod.MultiDictionary
List = LinkMethod.List
MultiList = LinkMethod.MultiList
Object = LinkMethod.Object
MultiObject = LinkMethod.MultiObject
FormattedText = LinkMethod.FormattedText

__all__ = [
'DirectLink',
'Dictionary',
'MultiDictionary',
'List',
'MultiList',
'Object',
'MultiObject',
'FormattedText'
]



class AttrLinkerException(RuntimeError):
    '''Base Exception For Whole Package'''
    pass


class LinkerException(AttrLinkerException):
    '''Base Exception for Linker'''
    pass

class ManagerException(AttrLinkerException):
    '''Base Exception for LinkManager'''

class LinkerNotReady(LinkerException):
    '''Linker is not ready'''
    pass

class LinkerExists(ManagerException):
    '''Linker is found in the manager's hashmap of linkers'''
    pass

class LinkerNotFound(ManagerException):
    '''Linker is not found in the manager's hashmap of linkers'''
    pass



from .utils import DictUpdater
from .attrLinker import LinkManager

from typing import List, Dict, Union, Any

# TODO: Make presets on like, linking to a dictionary's value, list index x, manipulating numbers and strings, etc.
# WARNING: For iterative linking, DO NOT use lambda directly, use it like in multiLinkDictionary using linkDictionary. OR you would be facing an issue, where every targetVar assigned, gets the last link converter.

def linkDictionary(targetClass: type, sourceVar: str, targetVar: str, sourceDictKey: str = None, default: Any = None, enableSetter: Any = False, doc: str = '', **kw):
    '''
    Link an attribute on an instance of targetClass to the source attribute's key, where the source attribute type is dictionary.
    Example: linkDictionary(Foo, 'dictionary', 'bar', 'bar_key', default='default_value')
    'Foo.bar' Gets value by 'Foo.dictionary.get('bar_key', 'default_value')'

    Params
    ------
    targetClass:type is the class for the linking to be applied at
    sourceVar:str is the source variable name of an instance of the targetClass, which must be of type dict
    targetVar:str is the attribute name on the class to be linked to
    sourceDictKey:str is the key to access the source dictionary, if None, use targetVar instead to access the dictionary.
    default:Any is the default return value for dictionary.get
    enableSetter:bool is basically, whether you want the targetVar attribute to have read-only access or full-access to the variable. Defaults to False(read-only)
    doc:str is the documentation string for the property created
    
    Extra keyword argument passed, would be passed directly to manager.bind
    '''
    if sourceDictKey is None:
        sourceDictKey = targetVar

    getterConverter = lambda dict: dict.get(sourceDictKey, default)
    setterConverter = lambda linkedSelf, replacement: DictUpdater(linkedSelf.__getattribute__(sourceVar), {sourceDictKey:replacement})
    manager = LinkManager._getDefault()
    manager.bind(targetClass, sourceVar, targetVar, getterConverter, setterConverter, setupOptions={'enableSetter': enableSetter}, doc=doc, **kw)


def multiLinkDictionary(targetClass: type, sourceVar: str, linkMap: Union[Dict[str,str], List[str]] = {}, **kw):
    '''
    Calls linkDictionary for every pair in linkMap.

    Params
    ------
    targetClass:type is the class for the linking to be applied at
    sourceVar:str is the source variable name of an instance of the targetClass, which must be of type dict
    linkMap:dict is the mapping for the linking, the mapping should be in the format as follows: {attribute_name_on_instance:key_in_dictionary,...} or [attribute_name_and_key_in_dict,...], where if you pass a list, it will generate the mapping from the list instead.

    Extra keyword argument passed, would be passed directly to linkDictionary
    '''
    if isinstance(linkMap, list):
        linkMap = {entry:entry for entry in linkMap} # Generate the dict mapping from list.
        
    for targetVar, sourceDictKey in linkMap.items():
        linkDictionary(targetClass, sourceVar, targetVar, sourceDictKey, **kw)


def formattedTextFromDict(targetClass: type, sourceVar: str, targetVar: str, formattableText: str, **kw):
    '''
    Creates a formatted text from given formattable_text which is formatted with the sourceVar dictionary.
    Similiar to linkDictionary, except you specify the dictionary keys in the formattableText, and you could specify multiple keys. Also setter is not available here.
    Would incurr a KeyError if the key(s) in the formattable_text is not found in given source dictionary.
    Example: formattedTextFromDict(User, 'data', 'full_name', '{first_name} {last_name}')
    'User.full_name' value is '{first_name} {second_name}'.format(**User.data)
    
    Params
    ------
    targetClass:type is the class for the linking to be applied at
    sourceVar:str is the source variable name of an instance of the targetClass, which must be of type dict
    targetVar:str is the attribute name on the class to be linked to
    formattableText:str is the text to be formatted with the data from the source. the format of the formattableText is "{key} {key2}" and so on.
    
    Extra keyword argument passed, would be passed directly to manager.bind
    '''
    getterConverter = lambda dict: formattableText.format(**dict)
    manager = LinkManager._getDefault()
    manager.bind(targetClass, sourceVar, targetVar, getterConverter, setupOptions={'enableSetter':False}, **kw)


def linkList(targetClass: type, sourceVar: str, targetVar: str, sourceIndex: int, enableSetter: bool = False, doc: str = '', **kw):
    '''
    Link an attribute on an instance of targetClass to source attribute's item on given index. 
    Example: linkList(Foo, '_list_of_bars', 'first_bar', 0)
    'Foo.first_bar' Gets its value by 'Foo._list_of_bars[0]'

    Params
    ------
    targetClass:type is the class for the linking to be applied at
    sourceVar:str is the source variable name of an instance of the targetClass
    targetVar:str is the attribute name on the class to be linked to
    sourceIndex:int is the target index on the source object
    enableSetter:bool is basically, whether you want the targetVar attribute to have read-only access or full-access to the variable. Defaults to False(read-only)
    doc:str is the documentation string for the property created
    
    Extra keyword argument passed, would be passed directly to manager.bind
    '''
    getterConverter = lambda _lst: _lst[sourceIndex]
    setterOverrider = lambda linkedSelf, linkedVar, replacement: [linkedSelf.__getattribute__(linkedVar).pop(sourceIndex), linkedSelf.__getattribute__(linkedVar).insert(sourceIndex, replacement)]
    manager = LinkManager._getDefault()
    manager.bind(targetClass, sourceVar, targetVar, getterConverter, setterOverrider=setterOverrider, setupOptions={'enableSetter': enableSetter}, doc=doc, **kw)


def multiLinkList(targetClass: type, sourceVar: str, linkMap: Dict[str, int] = {}, **kw):
    '''
    Calls linkList for every pair in linkMap.

    Params
    ------
    targetClass:type is the class for the linking to be applied at
    sourceVar:str is the source variable name of an instance of the targetClass, which must be of type list
    linkMap:dict is the mapping for the linking, the mapping should be in the format as follows: {attribute_name_on_instance:index_in_list,...}

    Extra keyword argument passed, would be passed directly to linkList
    '''
    for targetVar, sourceListIndex in linkMap.items():
        linkList(targetClass, sourceVar, targetVar, sourceListIndex, **kw)


def linkObject(targetClass: type, sourceVar: str, targetVar: str, sourceAttribute: str = None, enableSetter: bool = False, doc: str = '', **kw):
    '''
    Link an attribute on an instance of targetClass to source attribute's attribute, where the source attribute type is Any.
    Example: linkObject(Foo, 'obj', 'obj_attr1', 'attr_1')
    'Foo.obj_attr1' Gets value by 'Foo.obj.attr_1'

    Params
    ------
    targetClass:type is the class for the linking to be applied at
    sourceVar:str is the source variable name of an instance of the targetClass
    targetVar:str is the attribute name on the class to be linked to
    sourceAttribute:str is the attribute name on the source object, if None, use targetVar instead to access the object.
    enableSetter:bool is basically, whether you want the targetVar attribute to have read-only access or full-access to the variable. Defaults to False(read-only)
    doc:str is the documentation string for the property created
    
    Extra keyword argument passed, would be passed directly to manager.bind
    '''
    if sourceAttribute is None:
        sourceAttribute = targetVar
    
    getterConverter = lambda obj: obj.__getattribute__(sourceAttribute)
    #setterConverter = lambda linkedSelf, replacement: (lambda obj: [setattr(obj, sourceAttribute, replacement), obj][-1])(linkedSelf.__getattribute__(sourceVar))
    setterOverrider = lambda linkedSelf, linkedVar, replacement: setattr(linkedSelf.__getattribute__(linkedVar), sourceAttribute, replacement)
    manager = LinkManager._getDefault()
    manager.bind(targetClass, sourceVar, targetVar, getterConverter, setterOverrider=setterOverrider, setupOptions={'enableSetter': enableSetter}, doc=doc, **kw)


def multiLinkObject(targetClass: type, sourceVar: str, linkMap: Union[Dict[str, str], List[str]] = {}, **kw):
    '''
    Calls linkObject for every pair in linkMap.

    Params
    ------
    targetClass:type is the class for the linking to be applied at
    sourceVar:str is the source variable name of an instance of the targetClass
    linkMap:dict is the mapping for the linking, the mapping should be in the format as follows: {attribute_name_on_instance:attribute_on_source_object,...} or [attribute_name_on_instance_and_source_object], where if you pass a list, it will generate the mapping from the list instead.

    Extra keyword argument passed, would be passed directly to linkList
    '''
    if isinstance(linkMap, list):
        linkMap = {entry:entry for entry in linkMap} # Generate the dict mapping from list.

    for targetVar, sourceObjectAttr in linkMap.items():
        linkList(targetClass, sourceVar, targetVar, sourceObjectAttr, **kw)



from .attrLinker import Linker, LinkManager
from .presets import linkDictionary, multiLinkDictionary, formattedTextFromDict, linkList, multiLinkList, linkObject, multiLinkObject
from .linkMethod import LinkMethod
from .preparedLink import PreparedLink
from .decorator import LinkedClass
from .exceptions import LinkerException, LinkerExists, LinkerNotFound, LinkerNotReady

